/*
  A test for SQLite multi-threading.

  (c) Mircea Neacsu 2024
*/
#include <iostream>
#include <iomanip>
#include <mlib/thread.h>
#include <mlib/stopwatch.h>
#include <mlib/sqlitepp.h>

using std::vector, std::cout;

//------------------- Configuration section -----------------------------------

//number of threads used
#define NTHREADS 8

//number of INSERT statements
#define NSTATEMENTS 1000

// Should inserts be wrapped in a transaction?
#define USE_TRANSACTION true

// Should each thread have a separate connection to database or should
// all threads share the same database connection
#define SHARE_CONNECTION false

//----------------- End of configuration section ------------------------------

#if USE_TRANSACTION
// threads cannot share connection if using transaction
#define SHARE_CONNECTION false
#endif

#if SHARE_CONNECTION
/*
  If connection sharing is enabled, SQLite must be in SERIALIZED mode.
  In serialized mode, API calls that use any SQLite database connection
  or any object derived from such a database connection can be made safely
  from multiple threads.
*/
#define MODEL SQLITE_CONFIG_SERIALIZED
#else
/*
  If each thread has its own connection, SQLite can be in MULTITHREADED mode.
  In this mode, SQLite can be safely used by multiple threads provided that no
  single database connection nor any object derived from database connection,
  such as a prepared statement, is used in two or more threads at the same time.
*/
#define MODEL SQLITE_CONFIG_MULTITHREAD
#endif

// Thread that runs an SQL INSERT statement
class Inserter : public mlib::thread
{
public:
  Inserter(mlib::Database& db, bool transact_on)
    : q(db, "INSERT INTO tab VALUES (DATETIME('now', 'localtime'), ?1, ?2, ?3)")
    , transact (transact_on)
  {}

private:
  void run();
  mlib::Query q;
  bool transact;
};

void Inserter::run()
{
  mlib::stopwatch timer;
  mlib::erc result;
  mlib::Database& db = q.connection();

  timer.start();

  if (transact)
    db.exec("BEGIN TRANSACTION");

  for (int i = 0; i < NSTATEMENTS; i++)
  {
    int retries = 0;
    // bind initial values to each column
    q.bind(1, i);
    q.bind(2, (int)id());
    q.bind(3, retries);

    //try to add row
    while ((result = q.step()) == SQLITE_BUSY)
    {
      //if DB is busy...
      Sleep(0);                 //... wait a bit...
      q.reset().deactivate();   //... return statement to ready state....
      retries++;                //... increment number of retries...
      q.bind(3, retries);       //... and update the bound value
    }
    if (result != SQLITE_DONE)
    {
      cout << "Thread " << id() << " abnormal exit (" << result << ")\n";
      break;
    }
    q.reset();
  }
  if (transact)
  {
    while (db.exec("END TRANSACTION") == SQLITE_BUSY)
      Sleep(0);
  }
  timer.stop();
  cout << "Thread " << id() << " finished in " << std::setprecision(5) << timer.msecEnd()/1000. << " sec\n";
}

int main()
{
  vector<Inserter*> threads;
  vector<mlib::Database> connections;

  cout << "SQLITE version is: " << sqlite3_version << std::endl;
  cout << "Compile time threading supported: " << (sqlite3_threadsafe()?"yes":"no") << std::endl;
  
  cout << "Setting threading model to "
    << ((MODEL==SQLITE_CONFIG_MULTITHREAD)?"SQLITE_CONFIG_MULTITHREAD":"SQLITE_CONFIG_SERIALIZED")
    << std::endl;
  sqlite3_config(MODEL);
  
  mlib::Database db("test.db");
  db.exec("CREATE TABLE IF NOT EXISTS tab (time DATETIME, count NUMBER, thread NUMBER, retries NUMBER);\
               DELETE FROM tab;");

#if SHARE_CONNECTION
  /* If sharing connection, this is the one connection used by all threads*/
  mlib::Database connection("test.db");
#endif
  for (int i = 0; i < NTHREADS; i++)
  {
#if !SHARE_CONNECTION
    /* If not sharing connection, create a new connection for each thread */
    mlib::Database connection("test.db");
    connections.push_back(connection);
#endif
    threads.emplace_back(new Inserter(connection, USE_TRANSACTION));
  }

  for (int i = 0; i < NTHREADS; i++)
    threads[i]->start();

  mlib::wait_all(threads.data(), NTHREADS);

  //Do some statistics when finished
  mlib::Query stats(db, "SELECT MAX(retries), SUM(retries), COUNT(time) FROM tab");
  stats.step();

  cout << "\nTotal rows: " << stats.column_int(2) << std::endl;
  cout << "Total retries: " << stats.column_int(1) << " max retries: " << stats.column_int(0) << std::endl;

  return 0;
}

